#!/usr/local/cpanel/3rdparty/bin/perl

use strict;
use warnings;
no warnings 'uninitialized', 'once';
use LWP::UserAgent;
use IO::File;
use JSON;
use DBI;
use FindBin qw( $RealBin );
use Getopt::Std;
use Sys::Syslog qw(:macros :standard);

$|++;

our ($VERSION) = "2.13";
our ($PROGRAM) = 'PCCC Raptor FAST for Cpanel';
our ($COPYRIGHT) = 'Copyright (C) 2020 Peregrine Computer Consultants Corporation. All rights reserved.';
our ($VERBOSE) = 0;

my (%FORM, $FORM);
my (@tok_dom, $token, $domains, $domains_processed);
my (@toks, @doms);
my $domain_dom;
my ($users, $jsonuser, $jsondomain, @jsonemail);
my $i = 0;
my %opts;
my $testmode = 0;

$FORM = \%FORM;

getopts('d:vt', \%opts);

if(not defined $opts{'d'}) {
  $users = decode_json(<STDIN>);
  $jsonuser = $users->{'data'}->{'user'};
  $jsondomain = $users->{'data'}->{'args'}->{'email'};
  if(not defined $jsondomain) {
    $jsondomain = $users->{'data'}->{'args'}->{'address'};
  }
  # handle mailing lists
  if(not defined $jsondomain) {
    $jsondomain = $users->{'data'}->{'args'}->{'list'};
  }
} else {
  $jsondomain = $opts{'d'};
}

if($opts{'t'}) {
  $testmode = 1;
}

@jsonemail = split(/\@/, $jsondomain);
if(defined $jsonemail[1]) {
  $jsondomain = $jsonemail[1];
} elsif ( not defined $opts{'d'} ) {
  $jsondomain = $users->{'data'}->{'args'}->{'domain'};
}

if(not defined $jsondomain) {
  # be sure we have enough data to proceed
  openlog "Raptor-FAST", LOG_PID | LOG_PERROR, LOG_LOCAL7;
  syslog LOG_ERR, "invalid JSON data, user: %s", $jsonuser;
  closelog;
  print "1 invalid JSON data.";
  return (1, "invalid JSON data.");
}

print "$PROGRAM v$VERSION\n$COPYRIGHT\n\n" if $VERBOSE > 0;

=head1 NAME

raptor_fast_cpanel - Tool to send email users data from cPanel to Raptor FAST

=head1 SYNOPSIS

B<raptor_fast_cpanel>

=head1 DESCRIPTION

Tool to send email users data from cPanel to Raptor FAST, the software is automatically executed
when a new email address is created or modified in cPanel web gui.

The software reads a json file from C<STDIN>, parse it, and uses two config files,
C<config> with general options and C<config.token> with Raptor FAST tokens.

The json file on C<STDIN> can be bypassed with the -d parameter which specifies the domain name to be used.

When using the -v parameter some debug info will be printed on C<STDOUT>.

If the parameter -t is used there will be no submission to Raptor FAST and the json file will
be printed to C<STDOUT>.

An example of json file passed from cPanel follows:

C<data: {"result":"test+tstdmn.net","user":"tstdmn","args":
{"quota":"1024","email":test","domain":"tstdmn.net",
"password":"CHANGEMESOON","send_welcome_email":"0"}}>

=cut

if (-e "/etc/Raptor-FAST/config") {
  print "Loading Configuration... \n" if $VERBOSE > 0;
  require "/etc/Raptor-FAST/config";
  if (not (($main::token ne '') or ($main::tokenfile ne ''))) {
    print "Error.  You must have a valid Raptor FAST Token defined in your config file!\nExiting.\n\n";
    exit;
  }
} else {
  print "Error Loading Configuration! You must copy config.dist.cpanel to config and edit it for your installation.\n";
  exit;
}

#SANITIZE VARIABLES
$main::virtusertable ||= '/home/*/etc/*';
# $main::localhostnames ||= '/etc/mail/local-host-names';
$main::aliases ||= "/etc/valiases";
$main::domainaliases ||= "/etc/vdomainaliases";
$main::passwd ||= "passwd";

if(defined $opts{'v'}) {
  $main::VERBOSE = 1;
}

if ( -e "/etc/Raptor-FAST/config.token") {
  my $tokenfile = "/etc/Raptor-FAST/config.token";
  open my $tokens, $tokenfile or die "Could not open $tokenfile: $!";
  while( my $tokenline = <$tokens>)  {
    next if (/^#/ );
    if($tokenline !~ /^#/) {
      @tok_dom = split(/:/, $tokenline);
      $token = $tok_dom[0];
      $domains = $tok_dom[1];
      $domains_processed .= "," . $domains;
      push(@toks, $token);
      push(@doms, $domains);
      @tok_dom = undef;
    }
  }
  close $tokens;
  foreach my $token_dom (@toks) {
    $domain_dom = $doms[$i];
    chomp($domain_dom);
    chomp($jsondomain);
    if(($domain_dom eq "ALL") and ($i < $#toks)) {
      print "Error in config file config.token, 'ALL' domain must be on the last line.\n";
      exit;
    }
    if(($domain_dom eq "ALL") or ($domain_dom =~ /\Q$jsondomain\E/)) {
      if($domain_dom eq "ALL") {
        $jsonuser = undef;
      }
      &raptor_submit($FORM, token=>$token_dom, verbose=>$main::VERBOSE, virtusertable=>$main::virtusertable, domains=>$domain_dom, domains_processed=>$domains_processed, username=>$jsonuser);
    } elsif($domain_dom =~ /\Q$jsondomain\E/) {
      &raptor_submit($FORM, token=>$token_dom, verbose=>$main::VERBOSE, virtusertable=>$main::virtusertable, domains=>$domain_dom, domains_processed=>$domains_processed, username=>$jsonuser);
    }
    $i++;
  }
  if($i eq 0) {
    print "0 domain $jsondomain skipped";
  }
} else {
  @tok_dom = split(/:/, $main::token);
  $token = $tok_dom[0];
  $domains = $tok_dom[1];
  if(($domain_dom eq "ALL") or (index($jsondomain, $domain_dom) > 0)) {
    &raptor_submit($FORM, token=>$token, verbose=>$main::VERBOSE, virtusertable=>$main::virtusertable, domains=>$domains, username=>$jsonuser);
  }
}
exit;

sub raptor_submit {
  my ($FORM, %params) = @_;
  my ($fh, $fv, $valiasfile, $domainvaliasfile, $filename, %catchall_domains, %catchall_error_domains, $domain, @emails, $email, $s_emails, $name, %domains, @manual_domains, %catchall_redir_domains, $redir_domain);
  my @accounts = ();

  $fh = new IO::File;
  $fv = new IO::File;

  my @dir = glob $params{'virtusertable'};
  my %addresses;
  my $extra_emails = {};
  my $real_name;
  my $st_real_name;
  my ($conn, $subaccdb, $query);
  my @subaccrow = ();
  my @local_email;
  my @alias_emails;
  my $tmp_alias;
  my $primary_email;

  my $json = {accounts=>[], version=>$VERSION};

  # first member of glob could be not a directory
  if (( -f $dir[0]) or ( -d $dir[0])) {
    foreach my $path ( @dir ) {
      chomp($_);

      if ( ! -d $path ) {
        next;
      }
      my @domain = split("/", $path);

      $subaccdb = "/" . $domain[1] . "/" . $domain[2] . "/.subaccounts/storage.sqlite";

      # Consider only the username that has been modified in cPanel
      # if $params{'username'} a correct domain has been specified on the cmdline
      if((defined $params{'username'}) and ($domain[2] ne $params{'username'})) {
        next;
      }

      #ADD THE DOMAIN TO THE HASH OF ALL DOMAINS - FOR COMMENTS, ERRORS AND SPACE
      $domain = $domain[@domain-1];

      # skip domains not enabled in config file
      if(($params{'domains'} ne "ALL") and (index($params{'domains'}, $domain) < 0)) {
        next;
      }
      # skip domains previously processed
      # if there is only a token with "ALL" parameter, it should not be skipped
      if(($params{'domains'} eq "ALL") and (($params{'domains_processed'} !~ /^\,ALL/) and (index($params{'domains_processed'}, $domain) < 0))) {
        next;
      }
      $domains{$domain}++;
      print "DEBUG: Hash of All Domains in VirtUserTable $domain\n" if $VERBOSE > 0;

      $valiasfile = $main::aliases . "/" . $domain;
      my $extra_addr = undef;
      my $tmp_email;
      my $cnt = 0;

      $domainvaliasfile = $main::domainaliases . "/" . $domain;
      if ( (-f $domainvaliasfile) && ($fv->open("< $domainvaliasfile")) ) {
        while (<$fv>) {
          if ($_ =~ /(.*) : (.*)/) {
                    #THE DOMAIN REDIRECTS EMAIL - NEED TO CHECK IT AFTER THE PROCESSING IS DONE FOR VIRTUSERTABLE
                    $redir_domain = $2;
                    $catchall_redir_domains{$domain} = $redir_domain;
                    print "DEBUG: Catch-all with Redirection [$domain redirs to $redir_domain]\n" if $VERBOSE > 0;
                last;
          }
        }
      close($fv);
      }

      #CHECK DOMAINS WITH REDIRECTION IF THERE IS DOMAIN THEY REDIR TO IN ANOTHER LIST
      foreach $domain (keys %catchall_redir_domains) {

        print "DEBUG: ".($catchall_domains{$catchall_redir_domains{$domain}} + $catchall_error_domains{$catchall_redir_domains{$domain}})."\n" if $VERBOSE > 0;
        $catchall_domains{$domain}++;
        undef $catchall_redir_domains{$domain};
      }
      #ADD CATCHALL AS @domainname  AND CREATE HASH
      print "Catch-All without Error Domains:\n" if $VERBOSE > 0;
      foreach $domain (keys %catchall_domains) {
        print "\t$domain\n" if $VERBOSE > 0;
        push(@emails, "\@$domain");
      }
      print "\n" if $VERBOSE > 0;

      #READ /etc/passwd
      $filename = $path . "/" . $main::passwd;

      if ($fh->open("< $filename")) {
        while (<$fh>) {
          #IF NOT A COMMENT OR BLANK
          if ($_ !~ /(^#|^\s*$)/i) {
            $_ = m/^(.+?)\:/;
            $email = $1;
            $addresses{"$email\@$domain"}++;
            push (@emails, "$email\@$domain");
            push(@{$extra_emails->{"$email\@$domain"}}, "$email\@$domain");
            print "DEBUG: $email\n" if $VERBOSE > 0;
            if( -f $subaccdb ) {
              $conn = DBI->connect("DBI:SQLite:database=$subaccdb",{RaiseError => 1});
              $query = $conn->prepare("SELECT real_name FROM users WHERE username=?");
              $query->execute($email);
              @subaccrow = $query->fetchrow_array();
              $real_name->{"$email\@$domain"} = $subaccrow[0];
              $query->finish();
              $conn->disconnect();
            }
          }
        }
      close($fh);
      }
    push @manual_domains, $domain;
  }

  #ADD ADDRESSES + MANUAL DOMAINS TO EMAIL LIST
  print "Manual Domains (Relying on Aliases and Accounts or External Redirection):\n" if $VERBOSE > 0;
  foreach $domain (@manual_domains) {
    print "\t$domain\n" if $VERBOSE > 0;
  }

  #CHECK EACH EMAIL
  foreach $email (@emails) {
    $email =~ m/(.*)\@(.*)$/;
    $name = $1;
    $domain = $2;

    print "DEBUG: Domain for Email $email is $domain\n" if $VERBOSE > 0;
    print "DEBUG: Catchall_domains ".$catchall_domains{uc($domain)}."\n" if $VERBOSE > 0;

    unless ($catchall_domains{uc($domain)}) {
      $s_emails .= "$email\n";
      if(not defined $extra_emails->{$email}) {
        $extra_emails->{$email} = [ $email ];
      }
      if($real_name->{$email} eq "") {
         $st_real_name = $name;
      } else {
         $st_real_name = $real_name->{$email};
      }
      push(@accounts, {name=>$st_real_name, username=>$name, emails=>$extra_emails->{$email}, primary_email=>$email});
      $primary_email->{$email} = $email;
    }
   }

   if($params{'domains'} eq "ALL") {
     @dir = glob $main::aliases . "/*.*";
   } else {
     @dir = split(/,/, $params{'domains'});
   }
   foreach my $aliasdom ( @dir ) {
      $valiasfile = $aliasdom;
      if( ! -f $valiasfile ) {
        $valiasfile = $main::aliases . "/" . $valiasfile;
      }
      my @alpath = split(/\//, $valiasfile);
      $domain = pop @alpath;
      my $extra_addr = undef;
      my $tmp_email;
      my $cnt = 0;
      if ( (-f $valiasfile) && ($fv->open("< $valiasfile")) ) {
        while (<$fv>) {
          if ($_ =~ /\*: ("?):fail:/) {
             # Catch-all with error
             $catchall_error_domains{$domain}++;
             print "DEBUG: Catch-all with Error [$domain]\n" if $VERBOSE > 0;
             # last;
          } elsif ($_ =~ /\*: (.*)/) {
              my $mail_match = $1;
              my @a_extra_emails = split(/,/, $mail_match);
              for $extra_addr ( @a_extra_emails ) {
                if($extra_addr !~ /@/) {
                  if(length($extra_addr) > 2) {
                    $extra_addr = $mail_match . "@" . $domain;
                    push(@accounts, {name=>$mail_match, username=>$mail_match, emails=>[$extra_addr], primary_email=>$extra_addr});
                    # Catch-all
                    $extra_addr = "@" . $domain;
                    push(@accounts, {name=>$mail_match, username=>$mail_match, emails=>[$extra_addr], primary_email=>$extra_addr});
                  }
                } elsif(defined $extra_addr) {
                  push(@{$extra_emails->{$extra_addr}}, $extra_addr);
                }
              $cnt++;
              }
              if($cnt eq 0) {
                print "DEBUG: Catch-All for Domain without Error - Activated [due to $_]\n" if $VERBOSE > 0;
                #THE DOMAIN ALLOWS ALL EMAILS - IT HAS A CATCH-ALL MAKING ALL EMAILS VALID
                print "DEBUG: Catch-all without Error [$domain]\n" if $VERBOSE > 0;
                $catchall_domains{$domain}++;
                # last;
              }
          } elsif ($_ =~ /(.*): (.*mailman\s.*)/) {
            # Catch ml addresses
            @local_email = split(/\@/, $1);
            push(@accounts, {name=>$1, username=>$local_email[0], emails=>[$1], primary_email=>$1});
          } elsif ($_ =~ /(.*\@.*): (.*\@.*)/) {
             $tmp_email = $2;
             $tmp_alias = $1;
             if($domain =~ /\Q$tmp_email\E/) {
               @local_email = split(/\@/, $1);
               push(@accounts, {name=>$local_email[0], username=>$local_email[0], emails=>[$1], primary_email=>$1});
             } else {
               if($tmp_email =~ /\,/) {
                 @alias_emails = split(/,/, $tmp_email);
                 foreach my $alias_email ( @alias_emails ) {
                   push(@{$extra_emails->{$tmp_email}}, $alias_email);
                 }
                 if(exists($extra_emails->{$tmp_email}) and (scalar @{$extra_emails->{$tmp_email}} > 0)) {
                   # exclude autoresponders
                   if($tmp_email !~ /\s/) {
                     my %hash_emails   = map { $_, 1 } @{$extra_emails->{$tmp_email}};
                     my @unique_emails = keys %hash_emails;
                     push(@accounts, {name=>$tmp_alias, username=>$tmp_alias, emails=>\@unique_emails, primary_email=>$tmp_alias});
                   }
                 }
               } else {
                 # if the destination is in the same domain, add it as an extra email
                 # otherwise create a "fake" account
                 if(exists($primary_email->{$tmp_email})) {
                   push(@{$extra_emails->{$tmp_email}}, $tmp_alias);
                 } else {
                   @local_email = split(/\@/, $tmp_alias);
                   # exclude autoresponders
                   if($tmp_email !~ /\s/) {
                     push(@accounts, {name=>$local_email[0], username=>$local_email[0], emails=>[$tmp_email], primary_email=>$tmp_alias});
                   }
                 }
               }
             }
           } elsif ($_ =~ /(.*\@.*): :fail.*/) {
             print "DEBUG: email $1 with Error [$domain]\n" if $VERBOSE > 0;
           } else {
             # unreached code
           }
        }
      close($fv);
      }
   }

   # Save data into json struct
   push(@{$json->{'accounts'}}, @accounts);
   print "\n" if $VERBOSE > 0;
  }
  print "DEBUG:\n$s_emails\n" if $VERBOSE > 0;
  #REPORT THE RESULTS TO THE RAPTOR SERVER
  &submit_emails_to_raptor($FORM, json=>$json, token=>$params{'token'}, uri=>$main::uri);
}

sub submit_emails_to_raptor {
  my ($FORM, %params) = @_;
  my ($user_agent, $request, $result, $json_string, %emails, $key, $json);

  $json = $params{'json'};

  #REPORT THE RESULTS TO THE RAPTOR SERVER
  $user_agent = new LWP::UserAgent;
  $request = new HTTP::Request("POST", $params{'uri'});

  $request->content_type("application/x-www-form-urlencoded");

  #URL ENCODE THE INFORMATION
  $json_string = to_json($json, {pretty => 1});

  if($testmode eq 1) {
    print $json_string . "\n";
    exit;
  }
  $request->content("version=$VERSION&json_request=$json_string&token=$params{'token'}");
  $result = $user_agent->request($request);
  if ($result->is_error) {
    if(defined $users) {
      openlog "Raptor-FAST", LOG_PID | LOG_PERROR, LOG_LOCAL7;
      syslog LOG_ERR, "Error submitting JSON information to Raptor, user: %s", $jsonuser;
      closelog;
      print "1 Error submitting JSON information to Raptor.";
      return (1, "Error submitting JSON information to Raptor.");
    } else {
      openlog "Raptor-FAST", LOG_PID | LOG_PERROR, LOG_LOCAL7;
      syslog LOG_ERR, "Error submitting JSON information to Raptor, user: %s, error: %s", $jsonuser, $result->status_line;
      closelog;
      print "Error submitting JSON information to Raptor.\n";
      print "  Message: ".$result->status_line."\n";
      return (1, "Error submitting JSON information to Raptor.");
    }
  } else {
    print "0 Raptor submission successful";
    return (0, "Raptor submission successful");
  }
}
